@@ -8,7 +8,8 @@
 ##############################################################################
 import xdrlib, time, datetime, decimal, struct
 from firebirdsql.fberrmsgs import messages
-from firebirdsql import (DatabaseError, InternalError, OperationalError, 
+from firebirdsql import (DisconnectByPeer,
+    DatabaseError, InternalError, OperationalError,
     ProgrammingError, IntegrityError, DataError, NotSupportedError,
 )
 from firebirdsql.consts import *
@@ -40,15 +41,21 @@ def wire_operation(fn):
         return r
     return f
 
-def bytes_to_bint(b):           # Read as big endian
-    fmtmap = {1: 'b', 2: '>h', 4: '>l', 8: '>q'}
+def bytes_to_bint(b, u=False):           # Read as big endian
+    if u:
+        fmtmap = {1: 'B', 2: '>H', 4: '>L', 8: '>Q'}
+    else:
+        fmtmap = {1: 'b', 2: '>h', 4: '>l', 8: '>q'}
     fmt = fmtmap.get(len(b))
     if fmt is None:
         raise InternalError
     return struct.unpack(fmt, b)[0]
 
-def bytes_to_int(b):            # Read as little endian.
-    fmtmap = {1: 'b', 2: '<h', 4: '<l', 8: '<q'}
+def bytes_to_int(b, u=False):            # Read as little endian.
+    if u:
+        fmtmap = {1: 'B', 2: '<H', 4: '<L', 8: '<Q'}
+    else:
+        fmtmap = {1: 'b', 2: '<h', 4: '<l', 8: '<q'}
     fmt = fmtmap.get(len(b))
     if fmt is None:
         raise InternalError
@@ -159,16 +166,15 @@ class WireProtocol:
     buffer_length = 1024
 
     op_connect = 1
+    op_exit = 2
     op_accept = 3
+    op_disconnect = 6
     op_response = 9
-    op_dummy = 17
     op_attach = 19
     op_create = 20
     op_detach = 21
     op_transaction = 29
     op_commit = 30
-    op_commit_retaining = 50
-    op_rollback_retaining = 86
     op_rollback = 31
     op_open_blob = 35
     op_get_segment = 36
@@ -176,8 +182,14 @@ class WireProtocol:
     op_info_database = 40
     op_info_transaction = 42
     op_batch_segments = 44
-    op_allocate_statement = 62
+    op_que_events = 48
+    op_cancel_events = 49
+    op_commit_retaining = 50
+    op_event = 52
+    op_connect_request = 53
+    op_aux_connect = 53
     op_create_blob2 = 57
+    op_allocate_statement = 62
     op_execute = 63
     op_execute_immediate = 64
     op_fetch = 65
@@ -185,6 +197,7 @@ class WireProtocol:
     op_free_statement = 67
     op_prepare_statement = 68
     op_info_sql = 70
+    op_dummy = 71
     op_execute2 = 76
     op_sql_response = 78
     op_drop_database = 81
@@ -192,6 +205,7 @@ class WireProtocol:
     op_service_detach = 83
     op_service_info = 84
     op_service_start = 85
+    op_rollback_retaining = 86
 
     charset_map = {
     # DB CHAR SET NAME    :   PYTHON CODEC NAME (CANONICAL)
@@ -254,14 +268,7 @@ class WireProtocol:
         "convert bytes array to unicode string"
         return b.decode(self.charset_map.get(self.charset, self.charset))
 
-    def _parse_op_response(self):
-        b = recv_channel(self.sock, 16)
-        h = bytes_to_bint(b[0:4])         # Object handle
-        oid = b[4:12]                       # Object ID
-        buf_len = bytes_to_bint(b[12:])   # buffer length
-        buf = recv_channel(self.sock, buf_len, True)
-
-        # Parse status vector
+    def _parse_status_vector(self):
         sql_code = 0
         gds_codes = set()
         message = ''
@@ -286,11 +293,30 @@ class WireProtocol:
                 message = message.replace('@' + str(num_arg), n)
             n = bytes_to_bint(recv_channel(self.sock, 4))
 
+        return (gds_codes, sql_code, message)
+
+
+    def _parse_op_response(self):
+        b = recv_channel(self.sock, 16)
+        h = bytes_to_bint(b[0:4])         # Object handle
+        oid = b[4:12]                       # Object ID
+        buf_len = bytes_to_bint(b[12:])   # buffer length
+        buf = recv_channel(self.sock, buf_len, True)
+
+        (gds_codes, sql_code, message) = self._parse_status_vector()
         if sql_code or message:
             raise OperationalError(message, gds_codes, sql_code)
 
         return (h, oid, buf)
 
+    def _parse_op_event(self):
+        b = recv_channel(self.sock, 4096) # too large TODO: read step by step
+        # TODO: parse event name
+        db_handle = bytes_to_bint(b[0:4])
+        event_id = bytes_to_bint(b[-4:])
+
+        return (db_handle, event_id, {})
+
     @wire_operation
     def _op_connect(self):
         p = xdrlib.Packer()
@@ -455,7 +481,7 @@ class WireProtocol:
         send_channel(self.sock, p.get_buffer())
 
     @wire_operation
-    def _op_allocate_statement(self, trans_handle):
+    def _op_allocate_statement(self):
         p = xdrlib.Packer()
         p.pack_int(self.op_allocate_statement)
         p.pack_int(self.db_handle)
@@ -672,6 +698,62 @@ class WireProtocol:
         p.pack_int(blob_handle)
         send_channel(self.sock, p.get_buffer())
 
+    @wire_operation
+    def _op_que_events(self, event_names, ast, args, event_id):
+        params = bytes([1])
+        for name, n in event_names.items():
+            params += bytes([len(name)])
+            params += self.str_to_bytes(name)
+            params += int_to_bytes(n, 4)
+        p = xdrlib.Packer()
+        p.pack_int(self.op_que_events)
+        p.pack_int(self.db_handle)
+        p.pack_bytes(params)
+        p.pack_int(ast)
+        p.pack_int(args)
+        p.pack_int(event_id)
+        send_channel(self.sock, p.get_buffer())
+
+    @wire_operation
+    def _op_cancel_events(self, event_id):
+        p = xdrlib.Packer()
+        p.pack_int(self.op_cancel_events)
+        p.pack_int(self.db_handle)
+        p.pack_int(event_id)
+        send_channel(self.sock, p.get_buffer())
+
+    @wire_operation
+    def _op_connect_request(self):
+        p = xdrlib.Packer()
+        p.pack_int(self.op_connect_request)
+        p.pack_int(1)    # async
+        p.pack_int(self.db_handle)
+        p.pack_int(0)
+        send_channel(self.sock, p.get_buffer())
+
+        b = recv_channel(self.sock, 4)
+        while bytes_to_bint(b) == self.op_dummy:
+            b = recv_channel(self.sock, 4)
+        if bytes_to_bint(b) != self.op_response:
+            raise InternalError
+
+        h = bytes_to_bint(recv_channel(self.sock, 4))
+        recv_channel(self.sock, 8)  # garbase
+        ln = bytes_to_bint(recv_channel(self.sock, 4))
+        ln += ln % 4    # padding
+        family = bytes_to_bint(recv_channel(self.sock, 2))
+        port = bytes_to_bint(recv_channel(self.sock, 2), u=True)
+        b = recv_channel(self.sock, 4)
+        ip_address = '.'.join([str(byte_to_int(c)) for c in b])
+        ln -= 8
+        recv_channel(self.sock, ln)
+
+        (gds_codes, sql_code, message) = self._parse_status_vector()
+        if sql_code or message:
+            raise OperationalError(message, gds_codes, sql_code)
+
+        return (h, port, family, ip_address)
+
     @wire_operation
     def _op_response(self):
         b = recv_channel(self.sock, 4)
@@ -681,6 +763,19 @@ class WireProtocol:
             raise InternalError
         return self._parse_op_response()
 
+    @wire_operation
+    def _op_event(self):
+        b = recv_channel(self.sock, 4)
+        while bytes_to_bint(b) == self.op_dummy:
+            b = recv_channel(self.sock, 4)
+        if bytes_to_bint(b) == self.op_response:
+            return self._parse_op_response()
+        if bytes_to_bint(b) == self.op_exit or bytes_to_bint(b) == self.op_exit:
+            raise DisconnectByPeer
+        if bytes_to_bint(b) != self.op_event:
+            raise InternalError
+        return self._parse_op_event()
+
     @wire_operation
     def _op_sql_response(self, xsqlda):
         b = recv_channel(self.sock, 4)
@@ -709,3 +804,33 @@ class WireProtocol:
         b = recv_channel(self.sock, 32)     # ??? why 32 bytes skip
 
         return r
+
+    def _wait_for_event(self):
+        event_names = {}
+        event_id = 0
+        while True:
+            op = bytes_to_bint(recv_channel(self.sock, 4))
+            if op == self.op_dummy:
+                pass
+            elif op == self.op_exit or op == self.op_disconnect:
+                break
+            elif op == self.op_event:
+                db_handle = bytes_to_int(recv_channel(self.sock, 4))
+                ln = bytes_to_bint(recv_channel(self.sock, 4))
+                b = recv_channel(self.sock, ln, True)
+                assert byte_to_int(b[0]) == 1
+                i = 1
+                while i < len(b):
+                    ln = byte_to_int(b[i])
+                    s = self.connection.bytes_to_str(b[i+1:i+1+ln])
+                    n = bytes_to_int(b[i+1+ln:i+1+ln+4])
+                    event_names[s] = n
+                    i += ln + 5
+                recv_channel(self.sock, 8)  # ignore AST info
+
+                event_id = bytes_to_bint(recv_channel(self.sock, 4))
+                break
+            else:
+                raise InternalError
+
+        return (event_id, event_names)
